import itertools
import typing

import bs4


def edit_start(
    iterable,
    start_with_new: bool,
):
    if start_with_new:
        return iterable
    return iterable[::-1]


def limit_offset(
    iterable,
    offset: int,
    limit: typing.Optional[int],
) -> itertools.islice:
    return itertools.islice(
        itertools.islice(iterable, offset, None),
        limit,
    )


def urls_concat(
    urls: typing.List[typing.Any],
    sep: str = "",
) -> str:
    if urls == []:
        return ""

    full_url = urls.pop(0)
    for url in urls:
        url_with_sep = f"{sep}{url}"
        full_url += url_with_sep

    return full_url


def get_short_url(url: str) -> str:
    if url.startswith("http"):
        _, short_url = url.rsplit("/", maxsplit=1)
    else:
        short_url = url
    return short_url


def get_urls(
    tag: bs4.element.Tag,
    url_concat: str,
) -> typing.Tuple[str, str]:
    href: str = tag.a["href"]
    if href.startswith("/"):
        short_url = href[1:]  # without "/"
        url = urls_concat([url_concat, href])
    else:
        *_, short_url = href.rsplit("/", maxsplit=1)

        url = href
    return short_url, url



from typing import Optional

from pydantic import BaseModel, Field


class MangaBriefly(BaseModel):
    """
    This object represents a manga.
    """

    name_ru: str = Field(
        alias="rus_name",
        description="The manga's name in Russian",
    )
    name_en: str = Field(
        alias="en_name",
        description="The manga's name in English",
    )
    short_url: str = Field(
        alias="dir",
        description="The manga's short url",
    )
    url: str = Field(
        description="The manga's url",
    )
    id: int = Field(
        description="The manga's id",
    )
    photo: str = Field(
        alias="img",
        description="The manga's photo",
    )
    stars: float = Field(
        alias="avg_rating",
        description="The manga's rating",
    )
    year: Optional[int] = Field(
        alias="issue_year",
        description="Year of release of the manga",
    )
    count_chapters: int = Field(
        description="The number chapters of the manga",
    )



from typing import List, Optional

from pydantic import BaseModel, Field


class Chapter(BaseModel):
    """
    This object represents a chapter of the manga.
    """

    name: Optional[str] = Field(
        description="The chapter's name",
    )
    id: int = Field(
        description="The chapter's id",
    )
    tome: int = Field(
        description="The chapter of the volume",
    )
    chapter: float = Field(
        description="The chapter's numeric identifier, for example 23.3",
    )
    index: int = Field(
        description=(
            "The chapter's numeric value, "
            "for example index 2, is 1.1, because index 1 is chapter 1.0"
        ),
    )
    liked: int = Field(
        alias="score",
        description="The number of likes on chapter",
    )
    is_free: bool = Field(
        description="The chapter is free",
    )
    price: Optional[float] = Field(
        description=(
            "The chapter's price. "
            "The price may be even if the chapter is free"
        ),
    )
    free_date: Optional[str] = Field(
        alias="pub_date",
        description="The date when the chapter became or will be free"
    )
    date: str = Field(
        alias="upload_date",
        description="The chapter's release date",
    )
    translators: List[str] = Field(
        alias="publishers",
        description="The translators' names of the chapter",
    )



import json
import typing

import bs4

from .. import exceptions
from ..schemas import readmanga
from ..utils import edit_start, get_urls, limit_offset, urls_concat


class BaseUrls:
    """
    The base class of requests to the site https://readmanga.io
    """

    __hash__ = None
    __slots__ = ()

    URL_SITE = "https://readmanga.io"
    URL_SEARCH = "https://readmanga.io/search"
    URL_TRANSLATOR = "https://readmanga.io/list/person"
    URL_POPULAR = "https://readmanga.io/recommendations/top"
    URL_BEST = "https://readmanga.io/list?sortType=votes"
    URL_NEW = "https://readmanga.io/list?sortType=created"


class ReadMangaParser(BaseUrls):
    """
    The parser of site https://readmanga.io
    """

    __slots__ = ()

    def parse_manga_search(
        self,
        soup: bs4.BeautifulSoup,
        count: typing.Optional[int] = 1,
    ) -> typing.List[readmanga.MangaBriefly]:
        content = soup.find(class_="leftContent")\
                      .find(class_="tiles row")

        if content is None:
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "Check the correctness of the name that you passed. "
                "You may have specified too large offset! "
            ))

        manga = []
        for tag in limit_offset(
            iterable=content.find_all(class_="img"),
            offset=0,
            limit=count,
        ):
            tag: bs4.element.Tag

            name = tag.find_next(name="h3").a.string

            short_url, url = get_urls(tag, self.URL_SITE)

            photo = tag.a.find_next()["data-original"]

            manga.append(
                readmanga.MangaBriefly(
                    name=name,
                    short_url=short_url,
                    url=url,
                    photo=photo,
                )
            )

        return manga

    def parse_manga_info(
        self,
        soup: bs4.BeautifulSoup,
        short_url: str,
        chapters_offset: int = 0,
        chapters_count: typing.Optional[int] = None,
        chapters_primarily_new: bool = True,
    ) -> readmanga.Manga:
        manga = {}

        left_content = soup.find(class_="leftContent")

        # name, url, description
        for page_element in left_content.find(
            name=["h1", "meta"],
            recursive=False,
        ).find_next_siblings(
            name="meta",
            recursive=False,
        ):
            manga[page_element["itemprop"]] = page_element["content"]

        for photo in left_content.find(class_="picture-fotorama")\
                                 .find_all(
            name="img",
            recursive=False,
        ):
            manga.setdefault("photo", []).append(photo["data-full"])

        subject = left_content.find(class_="col-sm-7")

        # votes, stars
        type_by_name = {"ratingCount": int, "ratingValue": float}
        for rating in subject.find(class_="user-rating")\
                             .find_all(
            name="meta",
            recursive=False,
            itemprop=type_by_name.keys(),
        ):
            name = rating["itemprop"]

            manga[name] = type_by_name[name](rating["content"])

        type_info = subject.find(class_="subject-meta")

        try:
            _, manga["status"] = type_info.find(
                name="p",
                recursive=False,
            ).text.rsplit(maxsplit=1)
        except ValueError:
            manga["status"] = "завершен"

        manga["elem_year"] = int(
            type_info.find(class_="elem_year").a.string
        )

        category = type_info.find(class_="elem_category")
        if category:
            manga["elem_category"] = category.a.string

        # genres, tags, etc
        for tag in type_info.find_all(
            class_=["elem_genre", "elem_another", "elem_tag"],
        ):
            tag: bs4.element.Tag

            manga.setdefault("tags", []).append(tag.a.string)

        # authors, screenwriters, illustrators
        for name in ["elem_author", "elem_screenwriter", "elem_illustrator"]:
            for tag in type_info.find_all(class_=name):
                tag: bs4.element.Tag

                manga.setdefault(name, []).append(tag.a.string)

        # publishers
        for publisher in type_info.find_all(class_="elem_publisher"):
            publisher: bs4.element.Tag

            manga.setdefault("elem_publisher", []).append(
                publisher.string.strip(", "),
            )

        # translators
        for translator in type_info.find_all(class_="elem_translator"):
            translator: bs4.element.Tag

            name = translator.a.string

            translator_short_url = translator.a["href"]
            translator_url = urls_concat([
                self.URL_SITE, translator_short_url,
            ])

            manga.setdefault("elem_translator", []).append(
                readmanga.TranslatorBriefly(
                    name=name,
                    short_url=translator_short_url[1:],  # without "/"
                    url=translator_url,
                ),
            )

        # chapters
        subject = left_content.find(class_="table table-hover")
        if subject:
            chapters = subject.find(
                lambda tag: tag.parent.name != "thead",
                recursive=False,
            ).find_next_siblings()

            manga["count_chapters"] = len(chapters)

            for page_element in limit_offset(
                iterable=edit_start(
                    iterable=chapters,
                    start_with_new=chapters_primarily_new,
                ),
                offset=chapters_offset,
                limit=chapters_count,
            ):
                page_element: bs4.element.Tag

                if page_element.td.has_attr("rowspan"):
                    page_element.td.decompose()

                name = page_element.a.text.strip(" \n")

                chapter_url = urls_concat([
                    self.URL_SITE, page_element.a["href"],
                ])

                title = page_element.a["title"]
                if title:
                    translator, _ = title.rsplit(maxsplit=1)
                else:
                    translator = None

                date = page_element.td.find_next()\
                                      .find_next().string.strip(" \n")

                manga.setdefault("chapters", []).append(
                    readmanga.Chapter(
                        name=name,
                        url=chapter_url,
                        translator=translator,
                        date=date,
                    )
                )

        for similar in soup.find(class_="rightContent")\
                           .find(name="ol", class_="mangaList")\
                           .find_all(name="li"):
            similar: bs4.element.Tag

            name = similar.a.text.strip()

            similar_short_url, similar_url = get_urls(tag, self.URL_SITE)

            photo = similar.span["rel"]

            manga.setdefault("similar", []).append(
                readmanga.MangaBriefly(
                    name=name,
                    short_url=similar_short_url,
                    url=similar_url,
                    photo=photo,
                )
            )

        return readmanga.Manga(
            short_url=short_url,
            **manga,
        )

    def parse_chapter_pages(
        self,
        soup: bs4.BeautifulSoup,
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_first: bool = True,
    ) -> typing.List[readmanga.Page]:
        def check_script(tag: bs4.element.Tag) -> bool:
            if tag.name == "script":
                type = tag.get("type")
                if type is None or type != "text/javascript":
                    return False
                return tag.text.strip().startswith("var prevLink")

        script = soup.find(check_script)

        if script is None:
            raise exceptions.PagesNotFound((
                "The pages not found! "
                "Check the correctness of the url that you passed. "
            ))

        code = script.string

        pages_str = code[code.index("[["):code.index("]]") + 2]
        pages_json = json.loads(pages_str.replace("'", '"'))

        pages = []
        for page in limit_offset(
            iterable=edit_start(
                iterable=pages_json,
                start_with_new=primarily_first,
            ),
            offset=offset,
            limit=count,
        ):
            *url, width, height = page

            url = "".join(url)

            pages.append(
                readmanga.Page(
                    url=url,
                    width=width,
                    height=height,
                )
            )

        return pages

    def parse_translator_info(
        self,
        soup: bs4.BeautifulSoup,
        short_url: str,
        url: str,
    ) -> readmanga.Translator:
        translator = {}

        left_content = soup.find(class_="leftContent")

        # names
        for page_element in left_content.find(
            name="h1",
            recursive=False,
        ).find_all(
            name="span",
            recursive=False,
        ):
            page_element: bs4.element.Tag

            # name, eng-name, original-name = ...
            translator[page_element["class"][0]] = page_element.string

        subject = left_content.find(class_="flex-row")

        # votes, stars
        type_by_name = {"ratingCount": int, "ratingValue": float}
        for rating in subject.find(class_="user-rating")\
                             .find_all(
            name="meta",
            recursive=False,
            itemprop=type_by_name.keys(),
        ):
            name = rating["itemprop"]

            translator[name] = type_by_name[name](rating["content"])

        description = subject.find_next(class_="manga-description")
        if description:
            translator["manga-description"] = description.text.strip()

        date = subject.find(class_="subject-meta").find(
            name="p",
            recursive=False,
        )
        if date:
            translator["birthDate"] = date.time.string

        # vk / site / etc
        contact = left_content.find(target="_blank")
        if contact:
            translator["contact"] = urls_concat([
                self.URL_SITE, contact["href"],
            ])

        for work in left_content.find(class_="tiles row")\
                                .find_all(class_="img"):
            work: bs4.element.Tag

            name = work.find_next(name="h3").a.string

            work_short_url, work_url = get_urls(work, self.URL_SITE)

            photo = work.a.find_next()["data-original"]

            translator.setdefault("works", []).append(
                readmanga.MangaBriefly(
                    name=name,
                    short_url=work_short_url,
                    url=work_url,
                    photo=photo,
                )
            )

        return readmanga.Translator(
            short_url=short_url,
            url=url,
            **translator,
        )

    def parse_manga_popular(
        self,
        soup: bs4.BeautifulSoup,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        popular = soup.find(class_="leftContent")\
                      .find(class_="tiles row")\
                      .find_all(class_="img")

        if popular == []:
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "The source didn't return manga. "
            ))

        manga = []
        for popular in limit_offset(
            iterable=popular,
            offset=offset,
            limit=count,
        ):
            popular: bs4.element.Tag

            name = popular.find_next(name="h3").a.string

            short_url, url = get_urls(popular, self.URL_SITE)

            photo = popular.a.find_next()["data-original"]

            manga.append(
                readmanga.MangaBriefly(
                    name=name,
                    short_url=short_url,
                    url=url,
                    photo=photo,
                )
            )

        return manga

    def parse_manga_by_url(
        self,
        soup: bs4.BeautifulSoup,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        content = soup.find(class_="leftContent")\
                      .find(class_="tiles row")\
                      .find_all(class_="img")

        if content == []:
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "The source didn't return manga. "
            ))

        manga = []
        for manga_info in limit_offset(
            iterable=content,
            offset=0,
            limit=count,
        ):
            manga_info: bs4.element.Tag

            name = manga_info.find_next(name="h3").a.string

            short_url, url = get_urls(manga_info, self.URL_SITE)

            photo = manga_info.a.find_next()["data-original"]

            manga.append(
                readmanga.MangaBriefly(
                    name=name,
                    short_url=short_url,
                    url=url,
                    photo=photo,
                )
            )

        return manga



import typing

import bs4

from . import client


class Requests(client.Client):
    """
    The parent class for requests to various sites
    """

    __slots__ = ()

    def body(
        self,
        url: str,
        method: typing.Optional[str] = "GET",
        **kwargs,
    ) -> bs4.BeautifulSoup:
        page = self.request(url, method=method, **kwargs)

        return bs4.BeautifulSoup(page.content, "lxml")

    def json(
        self,
        url: str,
        method: typing.Optional[str] = "GET",
        **kwargs,
    ) -> typing.Dict[str, typing.Any]:
        page = self.request(url, method=method, **kwargs)

        return page.json()



from .chapter import Chapter
from .manga import Manga
from .manga_briefly import MangaBriefly
from .page import Page
from .translator import Translator
from .translator_briefly import TranslatorBriefly

__all__ = (
    "Chapter",
    "Manga",
    "MangaBriefly",
    "Page",
    "Translator",
    "TranslatorBriefly",
)



from typing import List, Optional

from pydantic import BaseModel, Field

from .chapter import Chapter
from .manga_briefly import MangaBriefly
from .translator_briefly import TranslatorBriefly


class Manga(BaseModel):
    """
    This object represents a manga.
    """

    name: str = Field(
        description="The manga's name",
    )
    short_url: str = Field(
        description="The manga's short url",
    )
    url: str = Field(
        description="The manga's url",
    )
    description: str = Field(
        description="The manga's description",
    )
    photo: List[str] = Field(
        description="Main the manga's photos",
    )
    votes: int = Field(
        alias="ratingCount",
        description="The number of votes to evaluate the manga",
    )
    stars: float = Field(
        alias="ratingValue",
        description="The manga's rating",
    )
    status: str = Field(
        description="The chapter's exit status",
    )
    year: int = Field(
        alias="elem_year",
        description="Year of release of the manga",
    )
    category: Optional[str] = Field(
        default=None,
        alias="elem_category",
        description="The manga's category (manga/manhua/manhwa)",
    )
    tags: List[str] = Field(
        description="Tags and genres of the manga",
    )
    authors: List[str] = Field(
        default=[],
        alias="elem_author",
        description="Authors of the manga",
    )
    screenwriters: List[str] = Field(
        default=[],
        alias="elem_screenwriter",
        description="Screenwriters of the manga",
    )
    illustrators: List[str] = Field(
        default=[],
        alias="elem_illustrator",
        description="Illustrators of the manga",
    )
    publishers: List[str] = Field(
        default=[],
        alias="elem_publisher",
        description="Publishers of the manga",
    )
    translators: List[TranslatorBriefly] = Field(
        default=[],
        alias="elem_translator",
        description="Translators of the manga",
    )
    chapters: List[Chapter] = Field(
        default=[],
        description="The manga's chapters",
    )
    count_chapters: int = Field(
        default=0,
        description="The number chapters of the manga in this branch",
    )
    similar: List[MangaBriefly] = Field(
        default=[],
        description="The similar manga",
    )



from pydantic import BaseModel, Field


class MangaBriefly(BaseModel):
    """
    This object represents a manga.
    """

    name: str = Field(
        description="The manga's name",
    )
    short_url: str = Field(
        description="The manga's short url",
    )
    url: str = Field(
        description="The manga's url",
    )
    photo: str = Field(
        description="The manga's photo",
    )



import typing

from .. import parsers, requests
from ..schemas import readmanga
from ..utils import get_short_url, urls_concat


class ReadManga(parsers.ReadMangaParser, requests.Requests):
    """
    The client for working with the site https://readmanga.io
    """

    __slots__ = ()

    def manga_search(
        self,
        name: str,
        offset: typing.Union[int, str] = 0,
        count: typing.Optional[int] = 1,
    ) -> typing.List[readmanga.MangaBriefly]:
        """
        Search for a similar manga by name.

        Returns a maximum of 50 results.

        * :name: The name to search for similar.
        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.parse_manga_search(
            soup=self.body(
                url=self.URL_SEARCH,
                method="POST",
                params={"q": name, "offset": offset},
            ),
            count=count,
        )

    def manga_info(
        self,
        url: str,
        chapters_offset: int = 0,
        chapters_count: typing.Optional[int] = None,
        chapters_primarily_new: bool = True,
    ) -> readmanga.Manga:
        """
        Returns the manga's information.

        * :url: Short url or full url.
         You can get short url from the object representing manga
         or just open any manga on the site
         and send what comes after the last slash. For example
         from "https://readmanga.live/dvorianstvo__A201945"
         need to pull out "dvorianstvo__A201945".
         Full url it's "https://readmanga.live/dvorianstvo__A201945".
        * :chapters_offset: The number of the manga chapter to start with.
        * :chapters_count: The number of the manga chapters to get.
        * :chapters_primarily_new: First of chapters is new.
        * :return: The object representing manga.
        """
        short_url = get_short_url(url)
        url = urls_concat([self.URL_SITE, short_url], "/")

        return self.parse_manga_info(
            soup=self.body(
                url=url,
            ),
            short_url=short_url,
            chapters_offset=chapters_offset,
            chapters_count=chapters_count,
            chapters_primarily_new=chapters_primarily_new,
        )

    def chapter_pages(
        self,
        url: str,
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_first: bool = True,
    ) -> typing.List[readmanga.Page]:
        """
        Returns the chapter's pages.

        Links to photos expire after a time.

        * :url:
         You can get the url from the object representing chapter
         or just open to the chapter any manga on the site
         and send full url. For example
         "https://readmanga.io/bashnia_boga__A339d2/vol1/0"
        * :offset: The number of the chapter pages to start with.
        * :count: The number of the chapter pages to get.
        * :primarily_first: First page is first.
        * :return: The objects representing page of the chapter.
        """
        return self.parse_chapter_pages(
            soup=self.body(
                url=url,
            ),
            offset=offset,
            count=count,
            primarily_first=primarily_first,
        )

    def translator_info(
        self,
        url: str,
    ) -> readmanga.Translator:
        """
        Returns the translator's information.

        * :url: Short url or full url.
         You can get short url from the object representing translator
         or just open to the page any translator on the site
         and send what comes after the last slash. For example
         from "https://readmanga.live/list/person/rikudou_sennin_clan"
         need to pull out "rikudou_sennin_clan". Full url it's
         "https://readmanga.live/list/person/rikudou_sennin_clan".
        * :return: The object representing translator
        """
        short_url = get_short_url(url)
        url = urls_concat([self.URL_TRANSLATOR, short_url], "/")

        return self.parse_translator_info(
            soup=self.body(
                url=url,
            ),
            short_url=short_url,
            url=url,
        )

    def manga_popular(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        """
        Returns information about a popular manga.

        Returns a maximum of 30 results
        (not by one request, but in general).

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.parse_manga_popular(
            soup=self.body(
                url=self.URL_POPULAR,
            ),
            offset=offset,
            count=count,
        )

    def manga_by_url(
        self,
        url: str,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        """
        Returns information about the manga by url.
        "manga_best" and "manga_new"
        are wrappers for this function.

        Returns a maximum of 70 results

        * :url: The url for returns the manga.
        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.parse_manga_by_url(
            soup=self.body(
                url=url,
                params={"offset": offset},
            ),
            count=count,
        )

    def manga_best(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        """
        Returns information about the best manga.

        Returns a maximum of 70 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_BEST,
            offset=offset,
            count=count,
        )

    def manga_new(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[readmanga.MangaBriefly]:
        """
        Returns information about a new manga.

        Returns a maximum of 70 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_NEW,
            offset=offset,
            count=count,
        )



class MangaNotFound(Exception):
    """
    If the manga was not found
    """
    pass


class MangaNeedAuthorization(Exception):
    """
    If you need authorization to view the manga
    """
    pass


class TranslatorNotFound(Exception):
    """
    If the translator was not found
    """
    pass


class ChaptersNotFound(Exception):
    """
    If the chapters was not found.
    Maybe everything is correct, but the chapters aren't found
    or the wrong branch is specified
    """
    pass


class ChapterNotFound(Exception):
    """
    If the chapter was not found
    """
    pass


class PagesNotFound(Exception):
    """
    If the pages was not found
    """
    pass


class IncorrectProtocol(Exception):
    """
    If the protocol incorrect
    """
    pass


class MethodNotFound(Exception):
    """
    If the method not found
    """
    pass


class BadStatusCode(Exception):
    """
    If the status code in the response indicates an error
    """
    pass


class PaymentRequired(Exception):
    """
    If the site returns a status code about the need for payment
    """
    pass


class ManyRequests(Exception):
    """
    If the site returns a status code about the many requests
    from your computer
    """
    pass



from typing import List, Optional

from pydantic import BaseModel, Field

from .translator_briefly import TranslatorBriefly


class Branch(BaseModel):
    """
    This object represents a branch.
    """

    id: int = Field(
        description="The branch's id",
    )
    photo: Optional[str] = Field(
        alias="img",
        description="The branch's photo",
    )
    translators: List[TranslatorBriefly] = Field(
        alias="publishers",
        description="The branch's translators",
    )
    liked: int = Field(
        alias="total_votes",
        description="The number of likes on chapters in this branch",
    )
    count_chapters: int = Field(
        description="The number chapters of the manga in this branch",
    )



import typing

import bs4

from .. import exceptions
from ..schemas import remanga
from ..utils import edit_start, limit_offset, urls_concat


class BaseUrls:
    """
    The base class of requests to the site https://remanga.org
    """

    __hash__ = None
    __slots__ = ()

    URL_SITE = "https://remanga.org"
    URL_SEARCH = "https://remanga.org/api/search"
    URL_MANGA = "https://remanga.org/manga"
    URL_MANGA_API = "https://remanga.org/api/titles"
    URL_CHAPTERS = "https://api.remanga.org/api/titles/chapters"
    URL_TEAM = "https://remanga.org/team"
    URL_TEAM_API = "https://remanga.org/api/publishers"
    URL_RECOMMENDATIONS = "https://remanga.org/api/titles/recommendations"
    URL_VIEWS = "https://remanga.org/api/titles/?ordering=-total_views"
    URL_LIKED = "https://remanga.org/api/titles/?ordering=-total_votes"
    URL_NEW = "https://remanga.org/api/titles/?ordering=-id"


class ReMangaParser(BaseUrls):
    """
    The parser of site https://remanga.org
    """

    __slots__ = ()

    def parse_manga_search(
        self,
        manga_json: typing.Dict[str, typing.Any],
    ) -> typing.List[remanga.MangaBriefly]:
        content = manga_json["content"]

        if content == []:
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "Check the correctness of the name that you passed. "
            ))

        manga = []
        for manga_info in content:
            manga_short_url = manga_info["dir"]
            manga_url = urls_concat([self.URL_MANGA, manga_short_url], sep="/")

            photo_url = urls_concat([self.URL_SITE, manga_info["img"]["high"]])

            manga.append(
                remanga.MangaBriefly(
                    rus_name=manga_info["rus_name"],
                    en_name=manga_info["en_name"],
                    dir=manga_short_url,
                    url=manga_url,
                    img=photo_url,
                    id=manga_info["id"],
                    avg_rating=float(manga_info["avg_rating"]),
                    issue_year=manga_info["issue_year"],
                    count_chapters=manga_info["count_chapters"],
                )
            )

        return manga

    def parse_manga_info(
        self,
        manga_json: typing.Dict[str, typing.Any],
        short_url: str,
    ) -> remanga.Manga:
        content = manga_json["content"]

        if content == []:
            message: str = manga_json["msg"]
            if message.startswith("Контент для взрослых"):
                raise exceptions.MangaNeedAuthorization((
                    "Needs authorization to view this manga! "
                    "The manga unavailable for viewing by minors. "
                    "You can change the headers and set token. "
                ))
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "Check the correctness of the short_url that you passed. "
            ))

        genres = [
            genre["name"]
            for genre in content["genres"]
        ]

        translators = []
        for translator in content["publishers"]:
            translator_short_url = translator["dir"]
            url = urls_concat(
                urls=[self.URL_TEAM, translator_short_url],
                sep="/",
            )

            translators.append(
                remanga.TranslatorBriefly(
                    name=translator["name"],
                    tagline=translator["tagline"],
                    dir=translator_short_url,
                    url=url,
                )
            )

        branches = []
        for branch_info in content["branches"]:
            branch_translators = []
            for translator in branch_info["publishers"]:
                translator_short_url = translator["dir"]
                url = urls_concat(
                    urls=[self.URL_TEAM, translator_short_url],
                    sep="/",
                )

                branch_translators.append(
                    remanga.TranslatorBriefly(
                        name=translator["name"],
                        tagline=translator["tagline"],
                        dir=translator_short_url,
                        url=url,
                    )
                )

            photo_short_url = branch_info["img"]
            if photo_short_url:
                photo_url = urls_concat([self.URL_SITE, photo_short_url])
            else:
                photo_url = None

            branches.append(
                remanga.Branch(
                    id=branch_info["id"],
                    img=photo_url,
                    publishers=translators,
                    total_votes=branch_info["total_votes"],
                    count_chapters=branch_info["count_chapters"],
                )
            )

        manga_url = urls_concat([self.URL_MANGA, short_url], sep="/")

        soup = bs4.BeautifulSoup(content["description"], "lxml")
        description = soup.text

        photo_url = urls_concat([self.URL_SITE, content["img"]["high"]])

        return remanga.Manga(
            rus_name=content["rus_name"],
            en_name=content["en_name"],
            dir=short_url,
            url=manga_url,
            id=content["id"],
            description=description,
            img=photo_url,
            count_rating=content["count_rating"],
            avg_rating=content["avg_rating"],
            count_bookmarks=content["count_bookmarks"],
            total_votes=content["total_votes"],
            total_views=content["total_views"],
            status=content["status"]["name"],
            issue_year=content["issue_year"],
            type=content["type"]["name"],
            genres=genres,
            publishers=translators,
            branches=branches,
            count_chapters=content["count_chapters"],
        )

    def parse_manga_chapters(
        self,
        chapters_json: typing.Dict[str, typing.Any],
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_new: bool = True,
    ) -> typing.List[remanga.Chapter]:
        content: list = chapters_json["content"]

        if content == []:
            raise exceptions.ChaptersNotFound((
                "The chapters was not found! "
                "Check the correctness of the branch_id that you passed. "
                "Maybe everything is correct, but the chapters aren't found. "
            ))

        chapters = []
        for chapter_info in limit_offset(
            iterable=edit_start(
                iterable=content,
                start_with_new=primarily_new,
            ),
            offset=offset,
            limit=count,
        ):
            translators = [
                translator["name"]
                for translator in chapter_info["publishers"]
            ]

            price = chapter_info["price"]
            if price:
                price = float(price)

            chapters.append(
                remanga.Chapter(
                    name=chapter_info["name"] or None,
                    id=chapter_info["id"],
                    tome=chapter_info["tome"],
                    chapter=chapter_info["chapter"],
                    index=chapter_info["index"],
                    score=chapter_info["score"],
                    is_free=not chapter_info["is_paid"],
                    price=price,
                    pub_date=chapter_info["pub_date"],
                    upload_date=chapter_info["upload_date"],
                    publishers=translators,
                )
            )

        return chapters

    def parse_chapter_pages(
        self,
        pages_json: typing.Dict[str, typing.Any],
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_first: bool = True,
    ) -> typing.List[remanga.Page]:
        content: typing.Dict[list, dict] = pages_json["content"]

        if content == []:
            raise exceptions.ChapterNotFound((
                "The chapter was not found! "
                "Check the correctness of the chapter_id that you passed. "
            ))

        pages = []
        for page in limit_offset(
            iterable=edit_start(
                iterable=content["pages"],
                start_with_new=primarily_first,
            ),
            offset=offset,
            limit=count,
        ):
            pages.append(
                remanga.Page(
                    link=page["link"],
                    height=page["height"],
                    width=page["width"],
                    count_comments=page["count_comments"],
                )
            )

        return pages

    def parse_translator_info(
        self,
        translator_json: typing.Dict[str, typing.Any],
        short_url: str,
    ) -> remanga.Translator:
        content: typing.Dict[list, dict] = translator_json["content"]

        if content == []:
            raise exceptions.TranslatorNotFound((
                "The translator not found! "
                "Check the correctness of the short_url that you passed. "
            ))

        contacts = {
            source: link
            for source, link in content["links"].items()
            if link != ""
        }

        soup = bs4.BeautifulSoup(content["description"], "lxml")
        description = soup.text

        photo_url = urls_concat([self.URL_SITE, content["img"]["high"]])

        translator_url = urls_concat([self.URL_TEAM, short_url], "/")

        return remanga.Translator(
            name=content["name"],
            tagline=content["tagline"],
            description=description,
            img=photo_url,
            dir=short_url,
            url=translator_url,
            rank=content["rank"]["name"],
            count_titles=content["count_titles"],
            count_votes=content["count_votes"],
            count_period_chapters=content["count_period_chapters"],
            links=contacts,
        )

    def parse_manga_by_url(
        self,
        manga_json: typing.Dict[str, typing.Any],
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        content = manga_json["content"]

        if content == []:
            raise exceptions.MangaNotFound((
                "The manga not found! "
                "The source didn't return manga. "
            ))

        manga = []
        for manga_info in limit_offset(
            iterable=content,
            offset=offset,
            limit=count,
        ):
            genres = manga_info["genres"]
            if genres is not None:
                genres = [
                    genre["name"]
                    for genre in manga_info["genres"]
                ]
            else:
                genres = []

            manga_short_url = manga_info["dir"]
            manga_url = urls_concat([self.URL_MANGA, manga_short_url], sep="/")

            photo_url = urls_concat([self.URL_SITE, manga_info["img"]["high"]])

            manga.append(
                remanga.MangaAverage(
                    rus_name=manga_info["rus_name"],
                    en_name=manga_info["en_name"],
                    dir=manga_short_url,
                    url=manga_url,
                    id=manga_info["id"],
                    img=photo_url,
                    avg_rating=float(manga_info["avg_rating"]),
                    total_votes=manga_info["total_votes"],
                    total_views=manga_info["total_views"],
                    issue_year=manga_info["issue_year"],
                    type=manga_info["type"],
                    genres=genres,
                    count_chapters=manga_info["count_chapters"],
                )
            )

        return manga



from typing import Optional

from pydantic import BaseModel, Field


class Chapter(BaseModel):
    """
    This object represents a chapter of the manga.
    """

    name: str = Field(
        description="The chapter's name",
    )
    url: str = Field(
        description="The chapter's url",
    )
    translator: Optional[str] = Field(
        default=None,
        description="The translator's name",
    )
    date: str = Field(
        description="The chapter's release date",
    )



from pydantic import BaseModel, Field


class TranslatorBriefly(BaseModel):
    """
    This object represents a translator.
    """

    name: str = Field(
        description="The translator's name",
    )
    short_url: str = Field(
        description="The translator's short url",
    )
    url: str = Field(
        description="The translator's url",
    )



from setuptools import find_packages, setup

setup(
    name="manga_parser",
    description="Convenient manga parser!",
    version="0.2.0",
    url="https://github.com/Desiders/manga_parser",
    author="Desiders",
    license="Apache2",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
    ],
    project_urls={
        "Source": "https://github.com/Desiders/manga_parser",
    },
    packages=find_packages(include=['manga_parser', 'manga_parser.*']),
    install_requires=[
        'httpx>=0.19.0,<1.0.0',
        'beautifulsoup4>=4.9.3',
        'lxml>=4.6.3',
        'pydantic>=1.8.2,<2.0.0',
    ],
    python_requires=">=3.5",
)



import typing

from .. import parsers, requests
from ..schemas import remanga
from ..utils import get_short_url, urls_concat


class ReManga(parsers.ReMangaParser, requests.Requests):
    """
    The client for working with the site https://remanga.org
    """

    __slots__ = ()

    def manga_search(
        self,
        name: str,
        count: int = 1,
    ) -> typing.List[remanga.MangaBriefly]:
        """
        Search for a similar manga by name.

        * :name: The name to search for similar.
        * :count: The number of manga to get.
        * :return: The objects representing manga.
        """
        return self.parse_manga_search(
            manga_json=self.json(
                url=self.URL_SEARCH,
                params={"query": name, "count": count},
            ),
        )

    def manga_info(
        self,
        url: str,
    ) -> remanga.Manga:
        """
        Returns the manga's information.

        * :url: Short url or full url.
         You can get the short url from the object representing manga
         or just open any manga on the site
         and send what comes after the last slash. For example
         from "https://remanga.org/manga/the_beginning_after_the_end"
         need to pull out "the_beginning_after_the_end".
         Full url it's "https://remanga.org/manga/the_beginning_after_the_end".
        * :return: The object representing manga.
        """
        short_url = get_short_url(url)
        url = urls_concat([self.URL_MANGA_API, short_url], "/")

        return self.parse_manga_info(
            manga_json=self.json(
                url=url,
            ),
            short_url=short_url,
        )

    def manga_chapters(
        self,
        branch_id: typing.Union[int, str],
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_new: bool = True,
    ) -> typing.List[remanga.Chapter]:
        """
        Returns the chapters' information.

        * :branch_id:
         You can get the branch id from the object representing manga
         or just open to the branch any manga on the site
         and send what comes after "branch=". For example from
         "https://remanga.org/manga/tomb-raider?subpath=content&branch=3786"
         need to pull out "3786".
        * :offset: The number of the manga chapter to start with.
        * :count: The number of the manga chapters to get.
        * :primarily_new: First of chapters is new.
        * :return: The objects representing manga.
        """
        return self.parse_manga_chapters(
            chapters_json=self.json(
                url=self.URL_CHAPTERS,
                params={"branch_id": branch_id},
            ),
            offset=offset,
            count=count,
            primarily_new=primarily_new,
        )

    def chapter_pages(
        self,
        chapter_id: typing.Union[int, str],
        offset: int = 0,
        count: typing.Optional[int] = None,
        primarily_first: bool = True,
    ):
        """
        Returns the chapter's pages.

        * :chapter_id:
         You can get the chapter id from the object representing chapter
         or just open to the chapter any manga on the site
         and send what comes after the last slash and "ch". For example
         from "https://remanga.org/manga/blade_of_demon_destruction/ch542404"
         need to pull out "542404" or "ch542404".
        * :offset: The number of the chapter pages to start with.
        * :count: The number of the chapter pages to get.
        * :primarily_first: First page is first.
        * :return: The objects representing page of the chapter.
        """
        if isinstance(chapter_id, str) and chapter_id.startswith("ch"):
            _, chapter_id = chapter_id.split("ch")

        url = urls_concat([self.URL_CHAPTERS, chapter_id], "/")

        return self.parse_chapter_pages(
            pages_json=self.json(
                url=url,
            ),
            offset=offset,
            count=count,
            primarily_first=primarily_first,
        )

    def translator_info(
        self,
        url: str,
    ) -> remanga.Translator:
        """
        Returns the translator's information.

        * :url: Short url or full url.
         You can get the short url from the object representing translator
         or just open to the page any translator on the site
         and send what comes after the last slash. For example
         from "https://remanga.org/team/vesperum"
         need to pull out "vesperum".
         Full url it's "https://remanga.org/team/vesperum".
        * :return: The object representing translator.
        """
        short_url = get_short_url(url)
        url = urls_concat([self.URL_TEAM_API, short_url], "/")

        return self.parse_translator_info(
            translator_json=self.json(
                url=url,
            ),
            short_url=short_url,
        )

    def manga_by_url(
        self,
        url: str,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        """
        Returns information about the manga by url.
        "manga_recommendations", "manga_views"
        "manga_votes" and "manga_new"
        are wrappers for this function.

        Returns a maximum of 20 results

        * :url: The url for returns the manga.
        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        page = offset // 20

        return self.parse_manga_by_url(
            manga_json=self.json(
                url=url,
                params={"page": page + 1},
            ),
            offset=offset,
            count=count,
        )

    def manga_recommendations(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        """
        Returns information about a popular manga.

        Returns a maximum of 20 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_RECOMMENDATIONS,
            offset=offset,
            count=count,
        )

    def manga_views(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        """
        Returns information about the most viewed manga.

        Returns a maximum of 20 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_VIEWS,
            offset=offset,
            count=count,
        )

    def manga_liked(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        """
        Returns information about the most liked manga.

        Returns a maximum of 20 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_LIKED,
            offset=offset,
            count=count,
        )

    def manga_new(
        self,
        offset: int = 0,
        count: typing.Optional[int] = None,
    ) -> typing.List[remanga.MangaAverage]:
        """
        Returns information about a new manga.

        Returns a maximum of 20 results

        * :offset: The manga number to start with.
        * :count: The number of manga to get. None is everything.
        * :return: The objects representing manga.
        """
        return self.manga_by_url(
            url=self.URL_NEW,
            offset=offset,
            count=count,
        )



from typing import Dict

from pydantic import BaseModel, Field


class Translator(BaseModel):
    """
    This object represents a translator.
    """

    name: str = Field(
        description="The translator's name",
    )
    short_description: str = Field(
        alias="tagline",
        description="The translator's short description",
    )
    description: str = Field(
        description="The translator's description",
    )
    short_url: str = Field(
        alias="dir",
        description="The translator's short url",
    )
    url: str = Field(
        description="The translator's url",
    )
    photo: str = Field(
        alias="img",
        description="The translator's photo",
    )
    rank: str = Field(
        description="The translator's rank",
    )
    count_works: int = Field(
        alias="count_titles",
        description="The translator's count works",
    )
    liked: int = Field(
        alias="count_votes",
        description="The number of likes on the chapters of all works",
    )
    chapters_month: int = Field(
        alias="count_period_chapters",
        description="Average number of chapters per month",
    )
    contacts: Dict[str, str] = Field(
        alias="links",
        description=(
            "The translator's contacts, such as: "
            "vk, youtube, fb, twitter, insta, discord"
        ),
    )



from typing import List, Optional

from pydantic import BaseModel, Field

from .manga_briefly import MangaBriefly


class Translator(BaseModel):
    """
    This object represents a translator.
    """

    name: str = Field(
        description="The translator's name",
    )
    name_en: Optional[str] = Field(
        default=None,
        alias="eng-name",
        description="The translator's name in English",
    )
    original_name: Optional[str] = Field(
        default=None,
        alias="original-name",
        description="The translator's original name",
    )
    votes: int = Field(
        alias="ratingCount",
        description="The number of votes to evaluate the translator",
    )
    stars: float = Field(
        alias="ratingValue",
        description="The translator's rating",
    )
    description: Optional[str] = Field(
        default=None,
        alias="manga-description",
        description="The translator's description",
    )
    short_url: str = Field(
        description="The translator's short url",
    )
    url: str = Field(
        description="The translator's url",
    )
    date_create: Optional[str] = Field(
        default=None,
        alias="birthDate",
        description="Team creation date",
    )
    contact: Optional[str] = Field(
        default=None,
        description="The translator's contact",
    )
    works: List[MangaBriefly] = Field(
        default=[],
        description="The author's works",
    )



import typing

import httpx

from . import exceptions


class BaseClient(object):
    """
    The parent class of parsers of clients
    for sending requests to sites
    """

    __hash__ = None
    __slots__ = (
        "client_kwargs",
        "request_kwargs",
        "_client",
    )

    def __init__(
        self,
        client_kwargs: typing.Dict[str, typing.Any] = {},
        request_kwargs: typing.Dict[str, typing.Any] = {},
    ):
        """
        You can pass various parameters for requests and the client,
        for example, headers or cookies.

        * :client_kwargs: Parameters for client.
        * :request_kwargs: Parameters for requests.
        """
        self.client_kwargs = client_kwargs
        self.request_kwargs = request_kwargs
        self._client: typing.Optional[httpx.Client] = None


class Client(BaseClient):
    """
    The client for sending requests to sites
    """

    def __enter__(self):
        return self

    def new_client(self) -> httpx.Client:
        return httpx.Client(**self.client_kwargs)

    @property
    def client(self) -> httpx.Client:
        if self._client is None or self._client.is_closed:
            self._client = self.new_client()
        return self._client

    def close(self):
        if self._client is not None:
            self._client.close()

    def __exit__(self, exc, val, trace):
        self.close()

    def request(
        self,
        url: str,
        method: typing.Optional[str] = "GET",
        **kwargs,
    ) -> httpx.Response:
        kwargs = dict(
            url=url,
            **self.request_kwargs,
            **kwargs,
        )
        try:
            if method == "GET":
                response = self.client.get(**kwargs)
            elif method == "POST":
                response = self.client.post(**kwargs)
            else:
                raise exceptions.MethodNotFound((
                    "There is support for only "
                    "GET and POST methods! "
                ))
        except httpx.UnsupportedProtocol:
            raise exceptions.IncorrectProtocol((
                "Missing either an \"http://\" or "
                "\"https://\" protocol! "
                "Make sure that the url is entered correctly. "
            ))
        if response.status_code >= 400:
            if response.status_code == 401:
                raise exceptions.MangaNeedAuthorization((
                    "Needs authorization to view this manga! "
                    "The manga unavailable for viewing by minors. "
                    "You can change the headers and set token. "
                ))
            elif response.status_code == 402:
                raise exceptions.PaymentRequired((
                    "The site returns a status_code about "
                    "the need for payment! "
                    "You can change the headers. "
                ))
            elif response.status_code == 429:
                raise exceptions.ManyRequests((
                    "The site returns a status_code about "
                    "many requests from your computer! "
                ))
            raise exceptions.BadStatusCode((
                "The source returned the error status_code! "
                "You may have passed some parameters incorrectly. "
                f"More: {response}. "
            ))
        return response



from pydantic import BaseModel, Field


class Page(BaseModel):
    """
    This object represents a page of the chapter.
    """

    url: str = Field(
        description="The page's url",
    )
    width: int = Field(
        description="The page's width",
    )
    height: int = Field(
        description="The page's height",
    )
